@@ -1,6 +1,6 @@
 #!/usr/bin/env python3
 # This file is part of Xpra.
-# Copyright (C) 2011-2021 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
@@ -51,7 +51,7 @@ class TempFileContext:
 class TestAuth(unittest.TestCase):
 
     def a(self, name: str):
-        pmod = "xpra.server.auth"
+        pmod = "xpra.auth"
         auth_module = __import__(pmod, globals(), locals(), [name], 0)
         mod = getattr(auth_module, name, None)
         assert mod, f"cannot load {name} from {pmod}"
@@ -276,7 +276,7 @@ class TestAuth(unittest.TestCase):
         else:
             raise RuntimeError("invalid digest should raise a ValueError")
         a.password_filename = "./this-path-should-not-exist"
-        assert a.load_password_file() is None
+        assert not a.load_password_file()
         assert a.stat_password_filetime() == 0
         # inaccessible:
         if POSIX:
@@ -313,7 +313,7 @@ class TestAuth(unittest.TestCase):
             raise Exception("authentication with no data should have failed")
 
     def test_sqlite(self):
-        from xpra.server.auth.sqlite import main as sqlite_main
+        from xpra.auth.sqlite import main as sqlite_main
         filename = temp_filename("sqlite")
         password = "hello"
 
@@ -470,7 +470,7 @@ def main():
     else:
         set_default_level(logging.CRITICAL)
     try:
-        from xpra.server import auth
+        from xpra import auth
         assert auth
     except ImportError as e:
         print("non server build, skipping auth module test: %s" % e)
