@@ -1,14 +1,14 @@
 #!/usr/bin/env python3
 # This file is part of Xpra.
-# Copyright (C) 2018-2020 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2018 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
 import os
 import unittest
 
-from xpra.util import AdHocStruct, typedict
-from xpra.os_util import OSEnvContext
+from xpra.util.objects import typedict, AdHocStruct
+from xpra.util.env import OSEnvContext
 from unit.test_util import silence_info
 from unit.server.mixins.servermixintest_util import ServerMixinTest
 
@@ -20,10 +20,11 @@ class NetworkStateMixinTest(ServerMixinTest):
             os.environ["XPRA_PING_TIMEOUT"] = "1"
             from xpra.server.mixins import networkstate
             from xpra.server.source.networkstate import NetworkStateMixin
-            assert NetworkStateMixin.is_needed(typedict())
+            assert NetworkStateMixin.is_needed(typedict({"network-state" : True}))
             opts = AdHocStruct()
             opts.pings = 1
             opts.bandwidth_limit = "1Gbps"
+            opts.bandwidth_detection = False
             #the limit for all clients:
             capped_at = 1*1000*1000*1000    #=="1Gbps"
             with silence_info(networkstate):
@@ -49,6 +50,7 @@ class NetworkStateMixinTest(ServerMixinTest):
                     raise Exception("should not allow %s (%s) as connection-data" % (v, type(v)))
             with silence_info(networkstate, "bandwidthlog"):
                 self.handle_packet(("bandwidth-limit", 10*1024*1024))
+
             def get_limit():
                 return self.source.get_info().get("bandwidth-limit", {}).get("setting", 0)
             self.assertEqual(10*1024*1024, get_limit())
@@ -57,6 +59,7 @@ class NetworkStateMixinTest(ServerMixinTest):
             self.assertEqual(min(capped_at, networkstate.MAX_BANDWIDTH_LIMIT), get_limit())
             #test source:
             timeouts = []
+
             def timeout(*args):
                 timeouts.append(args)
             self.source.disconnect = timeout
