@@ -128,7 +128,9 @@ def do_get_user_conf_dirs(_uid) -> list[str]:
             dirs.append(os.path.join(x, "Xpra"))
     except Exception:
         # fallback to hardcoded:
-        dirs = ["/Library/Application Support/Xpra"]
+        from os import getuid
+        if getuid() == 0:
+            dirs = ["/Library/Application Support/Xpra"]
     dirs.append("~/.xpra")
     return dirs
 
@@ -229,7 +231,7 @@ def do_get_nodock_command() -> list[str]:
     # AppStore builds have script wrappers:
     helper = os.path.join(base, "Resources", "scripts", "Xpra")
     if not os.path.exists(helper):
-        helper = os.path.join(_get_helpers_dir(), "Xpra")
+        helper = os.path.join(base, "Helpers", "Xpra")
     if not os.path.exists(helper):
         # having a dock is still better than
         # trying to run a command that does not exist!
