@@ -1,18 +1,20 @@
-# -*- coding: utf-8 -*-
 # This file is part of Xpra.
-# Copyright (C) 2010-2023 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
-#pylint: disable-msg=E1101
+# pylint: disable-msg=E1101
 
-from typing import Dict, Any, Optional, Tuple
+from typing import Any
 
-from xpra.util import typedict
+from xpra.os_util import gi_import
+from xpra.util.objects import typedict
 from xpra.server.mixins.stub_server_mixin import StubServerMixin
 from xpra.server.source.windows import WindowsMixin
 from xpra.net.common import PacketType
 from xpra.log import Logger
 
+GLib = gi_import("GLib")
+
 log = Logger("window")
 focuslog = Logger("focus")
 metalog = Logger("metadata")
@@ -24,30 +26,32 @@ class WindowServer(StubServerMixin):
     """
     Mixin for servers that forward windows.
     """
+    PREFIX = "window"
 
     def __init__(self):
         # Window id 0 is reserved for "not a window"
         self._max_window_id = 1
-        self._window_to_id : Dict[Any,int] = {}
-        self._id_to_window : Dict[int,Any] = {}
+        self._window_to_id: dict[Any, int] = {}
+        self._id_to_window: dict[int, Any] = {}
         self.window_filters = []
         self.window_min_size = 0, 0
-        self.window_max_size = 2**15-1, 2**15-1
+        self.window_max_size = 2 ** 15 - 1, 2 ** 15 - 1
 
     def init(self, opts) -> None:
         def parse_window_size(v, default_value=(0, 0)):
             try:
-                #split on "," or "x":
+                # split on "," or "x":
                 pv = tuple(int(x.strip()) for x in v.replace(",", "x").split("x", 1))
-                assert len(pv)==2
+                assert len(pv) == 2
                 w, h = pv
-                if w<=0 or h<=0 or w>=32768 or h>=32768:
+                if w <= 0 or h <= 0 or w >= 32768 or h >= 32768:
                     raise ValueError(f"invalid window size {w}x{h}")
                 return w, h
             except Exception:
                 return default_value
+
         self.window_min_size = parse_window_size(opts.min_size, (0, 0))
-        self.window_max_size = parse_window_size(opts.max_size, (2**15-1, 2**15-1))
+        self.window_max_size = parse_window_size(opts.max_size, (2 ** 15 - 1, 2 ** 15 - 1))
         minw, minh = self.window_min_size
         maxw, maxh = self.window_max_size
         self.update_size_constraints(minw, minh, maxw, maxh)
@@ -59,52 +63,46 @@ class WindowServer(StubServerMixin):
     def cleanup(self) -> None:
         for window in tuple(self._window_to_id.keys()):
             window.unmanage()
-        #this can cause errors if we receive packets during shutdown:
-        #self._window_to_id = {}
-        #self._id_to_window = {}
-
+        # this can cause errors if we receive packets during shutdown:
+        # self._window_to_id = {}
+        # self._id_to_window = {}
 
     def reinit_window_encoders(self) -> None:
-        #any window mapped before the threaded init completed
-        #may need to re-initialize its list of encoders:
+        # any window mapped before the threaded init completed
+        # may need to re-initialize its list of encoders:
         log("reinit_window_encoders()")
         for ss in self._server_sources.values():
             if isinstance(ss, WindowsMixin):
                 ss.reinit_encoders()
 
-
     def last_client_exited(self) -> None:
         self._focus(None, 0, [])
 
+    def get_server_features(self, _source) -> dict[str, Any]:
+        return {}
 
-    def get_server_features(self, _source) -> Dict[str,Any]:
+    def get_info(self, _proto) -> dict[str, Any]:
         return {
-            "window_refresh_config" : True,     #v4 clients assume this is available
-            "window-filters"        : True,     #v4 clients assume this is available
-            }
+            "state": {
+                "windows": sum(int(window.is_managed()) for window in tuple(self._id_to_window.values())),
+            },
+            "filters": tuple((uuid, repr(f)) for uuid, f in self.window_filters),
+        }
 
-    def get_info(self, _proto) -> Dict[str,Any]:
-        return {
-            "state" : {
-                "windows" : sum(int(window.is_managed()) for window in tuple(self._id_to_window.values())),
-                },
-            "filters" : tuple((uuid,repr(f)) for uuid, f in self.window_filters),
-            }
-
-    def get_ui_info(self, _proto, _client_uuids=None, wids=None, *_args) -> Dict[str,Any]:
+    def get_ui_info(self, _proto, _client_uuids=None, wids=None, *_args) -> dict[str, Any]:
         """ info that must be collected from the UI thread
             (ie: things that query the display)
         """
-        return {"windows" : self.get_windows_info(wids)}
+        return {"windows": self.get_windows_info(wids)}
 
-
-    def parse_hello(self, ss, caps:typedict, send_ui) -> None:
+    def parse_hello(self, ss, caps: typedict, send_ui) -> None:
         if send_ui:
             self.parse_hello_ui_window_settings(ss, caps)
 
-    def parse_hello_ui_window_settings(self, ss, c:typedict) -> None:
-        pass
-
+    def parse_hello_ui_window_settings(self, ss, c: typedict) -> None:
+        """
+        this method is overriden by the `seamless` server to set frame extents
+        """
 
     def add_new_client(self, *_args) -> None:
         minw, minh = self.window_min_size
@@ -116,46 +114,40 @@ class WindowServer(StubServerMixin):
             cmaxw, cmaxh = ss.window_max_size
             minw = max(minw, cminw)
             minh = max(minh, cminh)
-            if cmaxw>0:
+            if cmaxw > 0:
                 maxw = min(maxw, cmaxw)
-            if cmaxh>0:
+            if cmaxh > 0:
                 maxh = min(maxh, cmaxh)
         maxw = max(1, maxw)
         maxh = max(1, maxh)
-        if minw>0 and minw>maxw:
+        if minw > 0 and minw > maxw:
             maxw = minw
-        if minh>0 and minh>maxh:
+        if minh > 0 and minh > maxh:
             maxh = minh
         self.update_size_constraints(minw, minh, maxw, maxh)
 
-    def update_size_constraints(self, minw:int, minh:int, maxw:int, maxh:int) -> None:
-        #subclasses may update the window models
+    def update_size_constraints(self, minw: int, minh: int, maxw: int, maxh: int) -> None:
+        # subclasses may update the window models
         pass
 
-
-    def send_initial_data(self, ss, caps, send_ui:bool, share_count:int) -> None:
+    def send_initial_data(self, ss, caps, send_ui: bool, share_count: int) -> None:
         if not send_ui:
             return
         if not isinstance(ss, WindowsMixin):
             return
-        self.send_initial_windows(ss, share_count>0)
-        self.send_initial_cursors(ss, share_count>0)
-
+        self.send_initial_windows(ss, share_count > 0)
 
     def is_shown(self, _window) -> bool:
         return True
 
-
     def get_window_id(self, _gdkwindow) -> int:
-        #the X11 server can find the model from the window's xid
+        # the X11 server can find the model from the window's xid
         return 0
 
-
     def reset_window_filters(self) -> None:
         self.window_filters = []
 
-
-    def get_windows_info(self, window_ids) -> Dict[int,Dict[str,Any]]:
+    def get_windows_info(self, window_ids) -> dict[int, dict[str, Any]]:
         info = {}
         for wid, window in self._id_to_window.items():
             if window_ids is not None and wid not in window_ids:
@@ -163,11 +155,11 @@ class WindowServer(StubServerMixin):
             info[wid] = self.get_window_info(window)
         return info
 
-    def get_window_info(self, window) -> Dict[str,Any]:
+    def get_window_info(self, window) -> dict[str, Any]:
         from xpra.server.window.metadata import make_window_metadata
         info = {}
         for prop in window.get_property_names():
-            if prop=="icons" or prop is None:
+            if prop == "icons" or prop is None:
                 continue
             metadata = make_window_metadata(window, prop, skip_defaults=False)
             info.update(metadata)
@@ -175,10 +167,10 @@ class WindowServer(StubServerMixin):
             metadata = make_window_metadata(window, prop, skip_defaults=False)
             info.update(metadata)
         info.update({
-             "override-redirect"    : window.is_OR(),
-             "tray"                 : window.is_tray(),
-             "size"                 : window.get_dimensions(),
-             })
+            "override-redirect": window.is_OR(),
+            "tray": window.is_tray(),
+            "size": window.get_dimensions(),
+        })
         wid = self._window_to_id.get(window)
         if wid:
             wprops = self.client_properties.get(wid)
@@ -190,12 +182,11 @@ class WindowServer(StubServerMixin):
         metalog("updating metadata on %s: %s", window, pspec)
         wid = self._window_to_id.get(window)
         if not wid:
-            return  #window is already gone
+            return  # window is already gone
         for ss in self._server_sources.values():
             if isinstance(ss, WindowsMixin):
                 ss.window_metadata(wid, window, pspec.name)
 
-
     def _remove_window(self, window) -> int:
         wid = self._window_to_id[window]
         log("remove_window: %s - %s", wid, window)
@@ -215,8 +206,8 @@ class WindowServer(StubServerMixin):
         self.do_add_new_window_common(wid, window)
         return wid
 
-    def do_add_new_window_common(self, wid:int, window) -> None:
-        self._max_window_id = max(self._max_window_id, wid+1)
+    def do_add_new_window_common(self, wid: int, window) -> None:
+        self._max_window_id = max(self._max_window_id, wid + 1)
         props = window.get_dynamic_property_names()
         metalog("add_new_window_common(%s) watching for dynamic properties: %s", window, props)
         for prop in props:
@@ -229,9 +220,9 @@ class WindowServer(StubServerMixin):
         for ss in tuple(self._server_sources.values()):
             if not isinstance(ss, WindowsMixin):
                 continue
-            wprops = self.client_properties.get(wid, {}).get(ss.uuid)
+            wprops = self.client_properties.get(wid, {}).get(ss.uuid, {})
             x, y, w, h = geometry
-            #adjust if the transient-for window is not mapped in the same place by the client we send to:
+            # adjust if the transient-for window is not mapped in the same place by the client we send to:
             if "transient-for" in window.get_property_names():
                 transient_for = window.get_property("transient-for")
                 if transient_for:
@@ -245,17 +236,17 @@ class WindowServer(StubServerMixin):
                         if mapped_at:
                             wx, wy = pos
                             cx, cy = mapped_at[:2]
-                            if wx!=cx or wy!=cy:
-                                dx, dy = cx-wx, cy-wy
+                            if wx != cx or wy != cy:
+                                dx, dy = cx - wx, cy - wy
                                 if dx or dy:
                                     geomlog("adjusting new window position for client window offset: %s", (dx, dy))
                                     x += dx
                                     y += dy
             ss.new_window(ptype, wid, window, x, y, w, h, wprops)
 
-    def _process_damage_sequence(self, proto, packet : PacketType) -> None:
+    def _process_damage_sequence(self, proto, packet: PacketType) -> None:
         packet_sequence, wid, width, height, decode_time = packet[1:6]
-        if len(packet)>=7:
+        if len(packet) >= 7:
             message = packet[6]
         else:
             message = ""
@@ -274,36 +265,35 @@ class WindowServer(StubServerMixin):
             if damage:
                 damage(wid, window, x, y, width, height, options)
 
-    def _process_buffer_refresh(self, proto, packet : PacketType) -> None:
+    def _process_buffer_refresh(self, proto, packet: PacketType) -> None:
         """ can be used for requesting a refresh, or tuning batch config, or both """
         wid, _, qual = packet[1:4]
-        if len(packet)>=6:
+        if len(packet) >= 6:
             options = typedict(packet[4])
             client_properties = packet[5]
         else:
             options = typedict({})
             client_properties = {}
-        if wid==-1:
+        if wid == -1:
             wid_windows = self._id_to_window
         elif wid in self._id_to_window:
-            wid_windows = {wid : self._id_to_window.get(wid)}
+            wid_windows = {wid: self._id_to_window.get(wid)}
         else:
-            #may have been destroyed since the request was made
+            # may have been destroyed since the request was made
             log("invalid window specified for refresh: %s", wid)
             return
         log("process_buffer_refresh for windows: %s options=%s, client_properties=%s",
             wid_windows, options, client_properties)
         batch_props = options.dictget("batch", {})
         if batch_props or client_properties:
-            #change batch config and/or client properties
+            # change batch config and/or client properties
             self.update_batch_config(proto, wid_windows, typedict(batch_props), client_properties)
-        #default to True for backwards compatibility:
+        # default to True for backwards compatibility:
         if options.get("refresh-now", True):
-            refresh_opts = {"quality"           : qual,
-                            "override_options"  : True}
+            refresh_opts = {"quality": qual,
+                            "override_options": True}
             self._refresh_windows(proto, wid_windows, refresh_opts)
 
-
     def update_batch_config(self, proto, wid_windows, batch_props, client_properties) -> None:
         ss = self.get_server_source(proto)
         if ss is None:
@@ -329,20 +319,20 @@ class WindowServer(StubServerMixin):
             ss.refresh(wid, window, opts)
 
     def _idle_refresh_all_windows(self, proto) -> None:
-        self.idle_add(self._refresh_windows, proto, self._id_to_window, {})
+        GLib.idle_add(self._refresh_windows, proto, self._id_to_window, {})
 
     def refresh_all_windows(self) -> None:
         for ss in tuple(self._server_sources.values()):
             if not isinstance(ss, WindowsMixin):
                 self.do_refresh_windows(ss, self._id_to_window)
 
-    def get_window_position(self, _window) -> Optional[Tuple[int,int]]:
-        #where the window is actually mapped on the server screen:
+    def get_window_position(self, _window) -> tuple[int, int] | None:
+        # where the window is actually mapped on the server screen:
         return None
 
-    def _window_mapped_at(self, proto, wid:int, window, coords=None) -> None:
-        #record where a window is mapped by a client
-        #(in order to support multiple clients and different offsets)
+    def _window_mapped_at(self, proto, wid: int, window, coords=None) -> None:
+        # record where a window is mapped by a client
+        # (in order to support multiple clients and different offsets)
         ss = self.get_server_source(proto)
         if not ss:
             return
@@ -351,89 +341,52 @@ class WindowServer(StubServerMixin):
         else:
             ss.unmap_window(wid, window)
 
-
-    def _process_map_window(self, proto, packet : PacketType) -> None:
+    def _process_map_window(self, proto, packet: PacketType) -> None:
         log.info("_process_map_window(%s, %s)", proto, packet)
 
-    def _process_unmap_window(self, proto, packet : PacketType) -> None:
+    def _process_unmap_window(self, proto, packet: PacketType) -> None:
         log.info("_process_unmap_window(%s, %s)", proto, packet)
 
-    def _process_close_window(self, proto, packet : PacketType) -> None:
+    def _process_close_window(self, proto, packet: PacketType) -> None:
         log.info("_process_close_window(%s, %s)", proto, packet)
 
-    def _process_configure_window(self, proto, packet : PacketType) -> None:
+    def _process_configure_window(self, proto, packet: PacketType) -> None:
         log.info("_process_configure_window(%s, %s)", proto, packet)
 
-    def _get_window_dict(self, wids) -> Dict[int,Any]:
-        wd = {}
-        for wid in wids:
-            window = self._id_to_window.get(wid)
-            if window:
-                wd[wid] = window
-        return wd
-
-    def _process_suspend(self, proto, packet : PacketType) -> None:
-        eventslog("suspend(%s)", packet[1:])
-        ui = bool(packet[1])
-        wd = self._get_window_dict(packet[2])
-        ss = self.get_server_source(proto)
-        if ss:
-            ss.suspend(ui, wd)
-
-    def _process_resume(self, proto, packet : PacketType) -> None:
-        eventslog("resume(%s)", packet[1:])
-        ui = bool(packet[1])
-        wd = self._get_window_dict(packet[2])
-        ss = self.get_server_source(proto)
-        if ss:
-            ss.resume(ui, wd)
-
-
     def send_initial_windows(self, ss, sharing=False) -> None:
         raise NotImplementedError()
 
-
-    def send_initial_cursors(self, ss, sharing=False) -> None:
-        pass
-
-
     ######################################################################
     # focus:
-    def _process_focus(self, proto, packet : PacketType) -> None:
+    def _process_focus(self, proto, packet: PacketType) -> None:
         if self.readonly:
             return
         wid = packet[1]
         focuslog("process_focus: wid=%s", wid)
-        if len(packet)>=3:
+        if len(packet) >= 3:
             modifiers = packet[2]
         else:
             modifiers = None
         ss = self.get_server_source(proto)
         if ss:
             self._focus(ss, wid, modifiers)
-            #if the client focused one of our windows, count this as a user event:
-            if wid>0:
+            # if the client focused one of our windows, count this as a user event:
+            if wid > 0:
                 ss.user_event()
 
     def _focus(self, _server_source, wid, modifiers) -> None:
         focuslog("_focus(%s,%s)", wid, modifiers)
 
     def get_focus(self) -> int:
-        #can be overridden by subclasses that do manage focus
-        #(ie: not shadow servers which only have a single window)
-        #default: no focus
+        # can be overridden by subclasses that do manage focus
+        # (ie: not shadow servers which only have a single window)
+        # default: no focus
         return -1
 
-
     def init_packet_handlers(self) -> None:
-        self.add_packet_handlers({
-            "map-window" :          self._process_map_window,
-            "unmap-window" :        self._process_unmap_window,
-            "configure-window" :    self._process_configure_window,
-            "close-window" :        self._process_close_window,
-            "focus" :               self._process_focus,
-            "damage-sequence" :     self._process_damage_sequence,
-            "buffer-refresh" :      self._process_buffer_refresh,
-            "suspend" :             self._process_suspend,
-            "resume" :              self._process_resume,
-            })
+        self.add_packets(
+            "map-window", "unmap-window",
+            "configure-window", "close-window",
+            "focus",
+            "damage-sequence", "buffer-refresh",
+            main_thread=True)
